import numpy
import bisect
import matplotlib.pyplot as plt
import vegas
from math import sqrt
from fit import pvalueFromChisq
from parallelize import ParallelFunction
import copy
import itertools
from itertools import izip


class GVar:
    """Represents a gaussian random variate with a mean and standard deviation.

    Args:
      m (float or any object with `mean` and `sdev` attributes): If `m` is a
        float it specifies the mean of the gaussian variate. Otherwise the
        `mean` member of m is taken to be the mean.
      sdev (float or None, optional): The standard deviation of the gaussian
        variate. If ``None`` the standard deviation is taken from ``m.sdev`` (if
        `m` has that attribute) or set to zero. Defaults to ``None``.

    Attributes:
      mean (float): mean of the gaussian variate.
      sdev (float): standard deviation of the gaussian variate.

    Note:
      `GVar` objects support addition, subtraction, multiplication and division
      with floats or other `GVar` objects. In the latter case errors are combined
      in quadrature.

    """

    def __init__(self, m, sdev=None):
        if hasattr(m, 'mean') and hasattr(m, 'sdev'):
            self.mean = m.mean
            self.sdev = sdev if sdev is not None else m.sdev
        else:
            self.mean = float(m)
            self.sdev = abs(float(sdev)) if sdev is not None else 0.0

    def __add__(self, g):
        if hasattr(g, 'mean') and hasattr(g, 'sdev'):
            return GVar(self.mean + g.mean, sqrt(self.sdev**2 + g.sdev**2))
        else:
            return GVar(self.mean + g, self.sdev)

    def __radd__(self, g):
        if hasattr(g, 'mean') and hasattr(g, 'sdev'):
            return GVar(self.mean + g.mean, sqrt(self.sdev**2 + g.sdev**2))
        else:
            return GVar(self.mean + g, self.sdev)

    def __sub__(self, g):
        if hasattr(g, 'mean') and hasattr(g, 'sdev'):
            return GVar(self.mean - g.mean, sqrt(self.sdev**2 + g.sdev**2))
        else:
            return GVar(self.mean - g.mean, self.sdev)

    def __rsub__(self, g):
        if hasattr(g, 'mean') and hasattr(g, 'sdev'):
            return GVar(g.mean - self.mean, sqrt(self.sdev**2 + g.sdev**2))
        else:
            return GVar(g.mean - self.mean, self.sdev)

    def __mul__(self, a):
        if hasattr(a, 'mean') and hasattr(a, 'sdev'):
            return GVar(self.mean*a.mean,
                        sqrt((self.mean*a.sdev)**2 + (self.sdev*a.mean)**2))
        else:
            return GVar(self.mean*a, self.sdev*a)

    def __rmul__(self, a):
        if hasattr(a, 'mean') and hasattr(a, 'sdev'):
            return GVar(self.mean*a.mean,
                        sqrt((self.mean*a.sdev)**2 + (self.sdev*a.mean)**2))
        else:
            return GVar(self.mean*a, self.sdev*a)

    def __div__(self, a):
        if hasattr(a, 'mean') and hasattr(a, 'sdev'):
            return GVar(self.mean/a.mean,
                        sqrt((self.sdev/a.mean)**2 +
                             (self.mean*a.sdev/a.mean**2)**2))
        else:
            return GVar(self.mean/a, self.sdev/a)

    def __rdiv__(self, a):
        if hasattr(a, 'mean') and hasattr(a, 'sdev'):
            return GVar(a.mean/self.mean,
                        sqrt((a.sdev/self.mean)**2 +
                             (a.mean*self.sdev/self.mean**2)**2))
        else:
            return GVar(a/self.mean, a*self.sdev/self.mean**2)

    def __pow__(self, a):
        return GVar(self.mean**a, abs(a*self.mean**(a-1)*self.sdev))

    def __repr__(self):
        return 'GVar('+repr(self.mean)+', '+repr(self.sdev)+')'

    def __str__(self):
        """Return string representation of ``self``.

        The representation is designed to show at least
        one digit of the mean and two digits of the standard deviation.
        For cases where mean and standard deviation are not
        too different in magnitude, the representation is of the
        form ``'mean(sdev)'``. When this is not possible, the string
        has the form ``'mean +- sdev'``.

        """
        # taken from gvar.GVar in gvar module (lsqfit distribution)
        def ndec(x, offset=2):
            ans = offset - numpy.log10(x)
            ans = int(ans)
            if ans > 0 and x * 10. ** ans >= [0.5, 9.5, 99.5][offset]:
                ans -= 1
            return 0 if ans < 0 else ans
        dv = abs(self.sdev)
        v = self.mean

        # special cases
        if dv == float('inf'):
            return '%g +- inf' % v
        elif v == 0 and (dv >= 1e5 or dv < 1e-4):
            if dv == 0:
                return '0(0)'
            else:
                ans = ("%.1e" % dv).split('e')
                return "0.0(" + ans[0] + ")e" + ans[1]
        elif v == 0:
            if dv >= 9.95:
                return '0(%.0f)' % dv
            elif dv >= 0.995:
                return '0.0(%.1f)' % dv
            else:
                ndecimal = ndec(dv)
                return '%.*f(%.0f)' % (ndecimal, v, dv * 10. ** ndecimal)
        elif dv == 0:
            ans = ('%g' % v).split('e')
            if len(ans) == 2:
                return ans[0] + "(0)e" + ans[1]
            else:
                return ans[0] + "(0)"
        elif dv < 1e-6 * abs(v) or dv > 1e4 * abs(v):
            return '%g +- %.2g' % (v, dv)
        elif abs(v) >= 1e6 or abs(v) < 1e-5:
            # exponential notation for large |self.mean|
            exponent = numpy.floor(numpy.log10(abs(v)))
            fac = 10.**exponent
            mantissa = str(self/fac)
            exponent = "e" + ("%.0e" % fac).split("e")[-1]
            return mantissa + exponent

        # normal cases
        if dv >= 9.95:
            if abs(v) >= 9.5:
                return '%.0f(%.0f)' % (v, dv)
            else:
                ndecimal = ndec(abs(v), offset=1)
                return '%.*f(%.*f)' % (ndecimal, v, ndecimal, dv)
        if dv >= 0.995:
            if abs(v) >= 0.95:
                return '%.1f(%.1f)' % (v, dv)
            else:
                ndecimal = ndec(abs(v), offset=1)
                return '%.*f(%.*f)' % (ndecimal, v, ndecimal, dv)
        else:
            ndecimal = max(ndec(abs(v), offset=1), ndec(dv))
            return '%.*f(%.0f)' % (ndecimal, v, dv * 10. ** ndecimal)


class IntegrationResult:
    """Summarizes the results of a vegas integration.

    Args:
      itnres (list of GVar, optional): list of :py:class:`~myFitter.GVar`
        objects (or anything that has attributes 'mean' and 'sdev')
        corresponding to the results of individual iterations. Defaults to an
        empty list.

    Attributes:
      mean: mean value of weighted average
      sdev: standard deviation of weighted average
      chisq: chi-square value of the results of individual iterations
      itnres: same as `itnres` keyword argument.

    Note:
      You can join several IntegrationResult instances by adding them. Here is
      an example which updates and IntegrationResult `result` after each
      iteration and prints information about the current state of the
      integration::

        f = ...
        integ = Integrator(...)

        results = IntegrationResult()
        print results.hdr()
        for i in range(10):
            results += integ(f, nitn=1, neval=500)
            print results.info()

    """
    def __init__(self, itnres=[]):
        self.mean = None
        self.sdev = None
        self.chisq = None
        self.qval = None
        self.itnres = [GVar(r) for r in itnres]
        self._update()

    def _update(self):
        weight = 0.0
        wsum = 0.0
        for i, res in enumerate(self.itnres):
            weight = weight + 1.0/res.sdev**2
            wsum = wsum + res/res.sdev**2
        if self.itnres:
            wavg = wsum/weight
            self.chisq = sum(((r.mean - wavg.mean)/r.sdev)**2 \
                                 for r in self.itnres)
            self.mean = wavg.mean
            self.sdev = wavg.sdev
            if len(self.itnres) > 1:
                self.qval = pvalueFromChisq(self.chisq, len(self.itnres)-1)
            else:
                self.qval = None
        else:
            self.mean = None
            self.sdev = None
            self.chisq = None
            self.qval = None

    def __iadd__(self, r):
        if hasattr(r, 'itnres'):
            self.itnres.extend(r.itnres)
        elif hasattr(r, 'mean') and hasattr(r, 'sdev'):
            self.itnres.append(GVar(r))
        else:
            raise ValueError('cannot add object '+str(r)+' to IntegrationResult')
        self._update()
        return self

    def __add__(self, r):
        res = self.copy()
        res += r
        return res

    def copy(self):
        """Return an independent copy.

        """
        return copy.deepcopy(self)

    def drop(self, n):
        """Remove the results of the first n iterations.

        Args:
          n (int): number of iterations to remove.

        """
        n = max(0, min(n, len(self.itnres)))
        self.itnres = self.itnres[n:]
        self._update()

    def hdr(self):
        """Return the header for a running summary of the integration as a
        string.

        """
        ans  ='itn                  integral               wgt average  chi2/dof     Q\n'
        ans +='-----------------------------------------------------------------------'
        return ans

    def info(self):
        """Return information about the last iteration and the current average
        as a string.

        Note:
          Use together with hdr().

        """
        if len(self.itnres) > 1:
            fmt  ='{0:>3d}  {1:>24s}  {2:>24s}  {3:>8.1f}  {4:>3.0f}%'
            return fmt.format(len(self.itnres),
                              str(self.itnres[-1]), str(GVar(self)),
                              self.chisq/(len(self.itnres)-1),
                              100.0*self.qval)
        elif len(self.itnres) == 1:
            fmt ='{0:>3d}  {1:>24s}  {2:>24s}'
            return fmt.format(1, str(self.itnres[-1]), str(GVar(self)))
        else:
            return ''

    def __str__(self):
        res = IntegrationResult()
        ans = self.hdr()+'\n'
        for ires in self.itnres:
            res += ires
            ans += res.info()+'\n'
        return ans


def integrationResults(n):
    """Return an array of empty IntegrationResult instances.

    Args:
      n: number of elements of the array.

    """
    return numpy.array([IntegrationResult() for i in range(n)])


class Integrator:
    """Simple wrapper around ``vegas.Integrator``.

    Args:
      domain (list): Integration domain. Each element of the list should be
        a list or tuple with two elements defining the lower and upper bound
        of the integration domain of the related variable.
      parallel (dict or None, optional): Parallelisation options. The ``dict``
        should hold valid keyword arguments for :py:class:`ParallelFunction`.
        Defaults to ``None``, in which case the integrand is not parallelised.
      drop (int, optional): Number of initial iterations to exclude from the
        weighted average. Defaults to 0.
      log (output stream or None, optional): Output stream for log messages.
        Defaults to ``None``, in which case no log messages are printed.
      **kwargs (optional): All other keyword arguments are passed down to
        the ``vegas.Integrator`` constructor.

    """
    def __init__(self, domain, **kwargs):
        if 'parallel' in kwargs:
            self._parallel = kwargs.pop('parallel')
        else:
            self._parallel = None
        if 'nitn' in kwargs:
            self._nitn = kwargs.pop('nitn')
        else:
            self._nitn = 1
        if 'drop' in kwargs:
            self._drop = kwargs.pop('drop')
        else:
            self._drop = 0
        if 'log' in kwargs:
            self._log = kwargs.pop('log')
        else:
            self._log = None
        self._integ = vegas.Integrator(domain, nitn=1, **kwargs)

    def __call__(self, f, **kwargs):
        """Carries out the integration.

        Args:
          f (callable): The integrand.
          **kwargs (optional): Additional keyword arguments are the same as in
            the constructor and override the values specified there. For the
            `parallel` keyword the dicts are merged.

        Returns:
          IntegrationResult: An object holding the results of the iterations
            and their weighted average. If the integrand returns a list or array
            the return value is a numpy array of
            :py:class:`~myFitter.IntegrationResult` objects.

        """
        if 'parallel' in kwargs:
            if self._parallel is not None and kwargs['parallel'] is not None:
                popts = self._parallel.copy()
                popts.update(kwargs.pop('parallel'))
            else:
                popts = kwargs.pop('parallel')
        else:
            popts = self._parallel

        if 'nitn' in kwargs:
            nitn = kwargs.pop('nitn')
        else:
            nitn = self._nitn

        if 'drop' in kwargs:
            drop = kwargs.pop('drop')
        else:
            drop = self._drop

        if 'log' in kwargs:
            logstream = kwargs.pop('log')
        else:
            logstream = self._log

        if popts is not None:
            f = ParallelFunction(f, **popts)

        ires = None
        if logstream is not None:
            logstream.write(IntegrationResult().hdr()+'\n')
            logstream.flush()
        for i in range(nitn):
            result = self._integ(f, **kwargs)
            if hasattr(result, '__len__'):
                nintegrands = len(result)
                if ires is None:
                    ires = integrationResults(nintegrands)
                for j in xrange(nintegrands):
                    ires[j] += GVar(result.itn_results[0][j])
                if logstream is not None and nintegrands > 0:
                    logstream.write(ires[0].info()+'\n')
                    logstream.flush()
                if i < drop:
                    for j in xrange(nintegrands):
                        ires[j].drop(1)
            else:
                if ires is None:
                    ires = IntegrationResult()
                ires += GVar(result.itn_results[0])
                if logstream is not None:
                    logstream.write(ires.info()+'\n')
                    logstream.flush()
                if i < drop:
                    ires.drop(1)

        return ires

    def config(self, **kwargs):
        """Change or read configuration options.

        Note:
          This is essentially the same as the ``vegas.Integrator.set()`` method.
          The configuration ``dict`` contains an additional keywords
          ``'parallel'``, ``'drop'``, and ``'log'``.

        Args:
          **kwargs (optional): Configuration options to set. All keywords
            accepted by ``vegas.Integator.set()`` as well as
            `parallel`, `drop`, and `log` are valid.

        Returns:
          dict: The old configuration.

        """
        result = {'parallel' : copy.deepcopy(self._parallel),
                  'nitn'     : self._nitn,
                  'drop'     : self._drop,
                  'log'      : self._log}
        if 'parallel' in kwargs:
            self._parallel = kwargs.pop('parallel')
        if 'nitn' in kwargs:
            self._nitn = kwargs.pop('nitn')
        if 'drop' in kwargs:
            self._drop = kwargs.pop('drop')
        if 'log' in kwargs:
            self._log = kwargs.pop('log')
        vegascfg = self._integ.set(**kwargs)
        vegascfg.pop('nitn', None)
        result.update(vegascfg)
        return result

    def random(self, neval=None, **kwargs):
        """Generate random points.

        Args:
          neval (int or None, optional): Maximum number of points yielded by
            the iterator. Defaults to ``None``, in which case the value of the
            `neval` setting is used. Note that the number of yielded points will
            typically be less than `neval` (usually about half).
          **kwargs (optional): All additional keyword arguments are passed down
            to ``vegas.Integrator.random``.

        Yields:
          x (numpy array): A random point in the integration domain.
          w (float): The weight of the point. The weights of all yielded points
            add up to 1.

        """
        if neval is not None:
            oldsettings = self._integ.set(neval=neval)
        else:
            oldsettings = self._integ.set()
        for item in self._integ.random(**kwargs):
            yield item
        self._integ.set(oldsettings)

    def fill(self, f, neval, *hist, **kwargs):
        """Fill histograms with data.

        Args:
          f (callable): function to integrate.
          neval (int): number of evaluations.
          *hist: tuples describing the histograms to fill.
          **kwargs (optional): Allowed keyword arguments are the same as for
            the constructor and can be used to override settings made there.

        Note:
          The function `f` must be an integrand function. It should take a
          single numpy array `x` as argument and return an array whose first
          element is the function value of the integrand. The other elements can
          be 'binned'. The `*hist` arguments must be of the form ``(j, h)``
          where ``h`` is a :py:class:`~myFitter.Histogram1D` or
          :py:class:`~myFitter.Histogram2D` instance and `j` can be a callable
          object, a tuple of integers or an integer.

          If h is a :py:class:`~myFitter.Histogram1D` instance and `j` is an
          integer each bin of `h` with boundaries `a` and `b` will be filled
          with a GVar object holding the estimate of the integral of ``f(x)[0]``
          over the set defined by ``a <= f(x)[j] < b``. Alternatively `j` can be
          a tuple of one integer. If `j` is a callable object and `h` a
          :py:class:`~myFitter.Histogram1D` instance the integration domain for
          each bin is defined by ``a <= j(x, f(x)) < b``.

          If h is a :py:class:`~myFitter.Histogram2D` instance and `j` is a
          tuple of two integers each bin of `h` with x-boundaries `ax` and `bx`
          and y-boundaries `ay` and `by` will contain an estimate of the
          integral of ``f(x)[0]`` over the domain with ``ax <= f(x)[j[0]] < bx``
          and ``ay <= f(x)[j[1]] < by``. If `j` is a callable it should return a
          tuple of two numbers and the integration domain is defined by ``ax <=
          j(x, f(x))[0] < bx`` and ``ay <= j(x, f(x))[1] < by``.

        Returns:
          GVar: The value (and error) of the integral.

        """

        kwargs['neval'] = neval
        oldsettings = self.config(**kwargs)

        if self._parallel is not None:
            iterator = self._integ.random_batch
            f = ParallelFunction(f, **self._parallel)
            self._integ.set(nhcube_batch=neval)
            parallel = True
        else:
            iterator = self._integ.random
            parallel = False

        count = 0
        wsum = 0.0
        f0sum = 0.0
        f02sum = 0.0
        fsums = [numpy.zeros(h.data.shape) for g, h in hist]
        f2sums = [numpy.zeros(h.data.shape) for g, h in hist]
        for xbatch, wgtbatch in iterator():
            if not parallel:
                fxbatch = [f(xbatch)]
                xbatch = [xbatch]
                wgtbatch = [wgtbatch]
            else:
                fxbatch = f(xbatch)
            for x, wgt, fx in itertools.izip(xbatch, wgtbatch, fxbatch):
                count += 1
                wsum += wgt
                if not hasattr(fx, '__len__'):
                    fx = [fx]
                fx = numpy.asarray(fx)

                f0sum += wgt*fx[0]
                f02sum += (wgt*fx[0])**2

                for (g, h), fsum, f2sum in zip(hist, fsums, f2sums):
                    p = None
                    try:
                        if hasattr(g, '__call__'):
                            if isinstance(h, Histogram2D):
                                p = h._getbin(*g(x, fx))
                            else:
                                p = h._getbin(g(x, fx))
                        elif hasattr(g, '__len__'):
                            p = h._getbin(*tuple(fx[i] for i in g))
                        else:
                            p = h._getbin(fx[g])
                    except KeyError:
                        pass

                    if p is not None:
                        fsum[p] += wgt*fx[0]
                        f2sum[p] += (wgt*fx[0])**2

        delta = count*f02sum - f0sum**2
        if delta < 0.0:
            integral = GVar(f0sum/wsum, 0.0)
        else:
            integral = GVar(f0sum/wsum, sqrt(delta/(count-1.0))/wsum)

        for (g, h), fsum, f2sum in zip(hist, fsums, f2sums):
            for p, v in numpy.ndenumerate(h.data):
                delta = count*f2sum[p] - fsum[p]**2
                if delta < 0.0:
                    h.data[p] = GVar(fsum[p]/wsum, 0.0)
                else:
                    h.data[p] = GVar(fsum[p]/wsum, sqrt(delta/(count-1.0))/wsum)

        self.config(**oldsettings)
        return integral


def _hdiv(a, b):
    try:
        res = a/b
    except ZeroDivisionError:
        return None
    return res


class Histogram2D:
    """Represents two-dimensional histograms.

    Args:
      xvals (list or None, optional): list of central x-values of the histogram
        bins. Derived from `xbins` if ``None`. Defaults to ``None``.
      yvals (list or None, optional): list of central y-values of the histogram
        bins. Derived from `ybins` if ``None``. Defaults to ``None``.
      xbins (list or None, optional): list of bin boundaries in x direction.
        Derived from `xvals` if ``None``. Defaults to ``None``.
      ybins (list or None, optional): list of bin boundaries in y direction.
        Derived from `yvals` if ``None``. Defaults to ``None``.
      data (2D array of objects or None, optional): data members (z-values) of
        the histogram. The first index corresponds to the y-axis.
        If set to ``None`` a suitably dimensioned array of ``None`` values
        is created. Defaults to ``None``.
      like (Histogram2D or None, optional): If not ``None`` the central values
        and bin boundaries of another Histogram2D instance are used. This
        supercedes any values specified for `xvals`, `yvals`, `xbins`, or
        `ybins`. The ``data`` member is not copied. Defaults to ``None``.

    Attributes:
      xvals (numpy array): list of central x-values of the histogram bins.
      yvals (numpy array): list of central y-values of the histogram bins.
      xbins (numpy array): list of bin boundaries in x direction.
      ybins (numpy array): list of bin boundaries in y direction.
      data (numpy array): data members (z-values) of the histogram.
        The first index corresponds to the x-axis.

    Note:
      Histogram2D supports element-wise addition, subtraction, multiplication,
      and division with other Histogram2D instances as long as their
      `xvals`, `yvals`, `xbins` and `ybins` members are exactly equal. Addition,
      subtraction, multiplication or division by a constant value for bins
      is also supported.

    """
    def __init__(self, xvals=None, yvals=None, xbins=None, ybins=None,
                 data=None, like=None):
        if like is not None:
            xvals = numpy.array(like.xvals)
            yvals = numpy.array(like.yvals)
            xbins = numpy.array(like.xbins)
            ybins = numpy.array(like.ybins)
        else:
            if xvals is None and xbins is None:
                raise TypeError('Either `like`, `x` or `xbins` must be set.')
            elif xvals is None:
                if not hasattr(xbins, '__len__'):
                    raise ValueError('Argument `xbins` must be iterable.')
                elif len(xbins) < 2:
                    raise ValueError('Argument `xbins` must have at least 2 entries.')
                xbins = numpy.array(xbins)
                xvals = 0.5*(xbins[1:] + xbins[:-1])
            else:
                if not hasattr(xvals, '__len__'):
                    raise ValueError('Argument `xvals` must be iterable.')
                elif len(xvals) < 2 and xbins is None:
                    raise ValueError('Argument `xvals` must have at least 2 entries or xbins must be given.')
                elif len(xvals) < 1:
                    ValueError('Argument `xvals` must have at least 1 entry.')
                xvals = numpy.array(xvals)
                if xbins is None:
                    xbins = 0.5*(xvals[1:] + xvals[:-1])
                    xbins = numpy.array([xvals[0]-(xbins[0]-xvals[0])] + \
                                            list(xbins) + \
                                            [xvals[-1]+(xvals[-1]-xbins[-1])])
                else:
                    if len(xbins) != len(xvals)+1:
                        raise ValueError('Argument `xbins` has wrong length.')
                    for i, x in enumerate(xvals):
                        if x < xbins[i]:
                            raise ValueError('Bin centre outside of bin.')
                    if xvals[-1] > xbins[-1]:
                            raise ValueError('Bin centre outside of bin.')

            if yvals is None and ybins is None:
                raise TypeError('Either `like`, `yvals` or `ybins` must be set.')
            elif yvals is None:
                if not hasattr(ybins, '__len__'):
                    raise ValueError('Argument `ybins` must be iterable.')
                elif len(ybins) < 2:
                    raise ValueError('Argument `ybins` must have at least 2 entries.')
                ybins = numpy.array(ybins)
                yvals = 0.5*(ybins[1:] + ybins[:-1])
            else:
                if not hasattr(yvals, '__len__'):
                    raise ValueError('Argument `yvals` must be iterable.')
                elif len(yvals) < 2 and ybins is None:
                    raise ValueError('Argument `yvals` must have at least 2 entries or `ybins` must be given.')
                elif len(yvals) < 1:
                    raise ValueError('Argument `yvals` must have at least 1 entry.')
                yvals = numpy.array(yvals)
                if ybins is None:
                    ybins = 0.5*(yvals[1:] + yvals[:-1])
                    ybins = numpy.array([yvals[0]-(ybins[0]-yvals[0])] + \
                                            list(ybins) + \
                                            [yvals[-1]+(yvals[-1]-ybins[-1])])
                else:
                    if len(ybins) != len(yvals)+1:
                        raise ValueError('Argument `xbins` has wrong length.')
                    for i, y in enumerate(yvals):
                        if y < ybins[i]:
                            raise ValueError('Bin centre outside of bin.')
                    if yvals[-1] > ybins[-1]:
                            raise ValueError('Bin centre outside of bin.')

        self.xvals = xvals
        self.yvals = yvals
        self.xbins = xbins
        self.ybins = ybins

        if data is None:
            self.data = numpy.empty((len(yvals), len(xvals)), dtype=object)
        else:
            if not hasattr(data, 'shape') or len(data.shape) != 2:
                raise TypeError('`data` argument must be 2D-array.')
            if len(yvals) != data.shape[0] or len(xvals) != data.shape[1]:
                raise ValueError('`data` argument has wrong dimensions.')
            self.data = data.copy()

    def copy(self):
        """Return an independent copy of self.

        """
        return Histogram2D(like=self, data=self.data)

    def _getbin(self, x, y):
        if y < self.ybins[0] or y > self.ybins[-1] or \
                x < self.xbins[0] or x > self.xbins[-1]:
            raise KeyError('Invalid key `'+repr((x, y))+'`.')
        py = max(bisect.bisect_left(self.ybins, y) - 1, 0)
        px = max(bisect.bisect_left(self.xbins, x) - 1, 0)
        return py, px

    def __getitem__(self, xy):
        """Return the data value of a specific bin

        Args:
          xy (tuple): A pair of coordinates.

        Returns:
          The data value of the bin which contains the coordinates `xy`.

        Raises:
          KeyError: The coordinates `xy` are outside the range of the histogram.

        """
        x, y = xy
        return self.data[self._getbin(x, y)]
        
    def __setitem__(self, xy, val):
        """Set the data value of a specific bin

        Args:
          xy (tuple): A pair of coordinates.
          val: A data value. The value of the bin containing the coordinates
            `xy` will be set to `val`.

        Raises:
          KeyError: The coordinates `xy` are outside the range of the histogram.

        """
        x, y = xy
        self.data[self._getbin(x, y)] = val

    def center(self, x, y):
        """Return the center of the bin to which (x,y) belongs.

        """
        py, px = self._getbin(x, y)
        return (self.xvals[px], self.yvals[py])        

    def lb(self, x, y):
        """Return the lower left corner of the bin to which (x,y) belongs.

        """
        py, px = self._getbin(x, y)
        return (self.xbins[px], self.ybins[py])

    def ub(self, x, y):
        """Return the upper right corner of the bin to which (x,y) belongs.

        """
        py, px = self._getbin(x, y)
        return (self.xbins[px+1], self.ybins[py+1])
        
    def __iter__(self):
        """Iterate over the bin centres.

        Yields:
          x (float): The x-coordinate of the bin centre.
          y (float): The y-coordinate of the bin centre.

        """
        for y in self.yvals:
            for x in self.xvals:
                yield (x, y)
    
    def iteritems(self):
        """Iterate over all elements of the histogram.

        Yields:
          tuple: ``((x, y), v)`` where ``(x, y)`` is the bin center and `v` the
            value associated with that bin.
        
        """
        for iy, y in enumerate(self.yvals):
            for ix, x in enumerate(self.xvals):
                yield (x, y, self.data[ix, iy])

    def max(self):
        """Return the largest data member in the histogram.

        Note:
          `None` values are ignored.

        """
        result = None
        for x, y, v in self.iteritems():
            if v is not None:
                result = v if result is None else max(v, result)
        return result

    def min(self):
        """Return the smallst data member in the histogram.

        Note:
          `None` values are ignored.

        """
        result = None
        for x, y, v in self.iteritems():
            if v is not None:
                result = v if result is None else min(v, result)
        return result
        
    def set(self, h):
        """Set all data values of the histogram.

        Args:
          h (Histogram2D or object): If `h` is a Histogram2D instance the
            function iterates over all bins of `h` and sets the corresponding
            bins of `self` to their values. Otherwise all bin values of
            `self` are set to `h`.

        """
        if isinstance(h, Histogram2D):
            for x, y, v in h.iteritems():
                self[x, y] = v
        else:
            self.data.fill(h)

    def settomin(self, h):
        """Set all bins to the minimum of their current value and h.

        Args:
          h (Histogram2D or object): If `h` is a Histogram2D instance the
            function iterates over all bins of `h` and sets the corresponding
            bins of `self` to the minimum of the two values. Otherwise all bins
            of `self` are set to the minimum of their current value and `h`.

        """
        if isinstance(h, Histogram2D):
            for x, y, v in h.iteritems():
                p = self._getbin(x,y)
                s = self.data[p]
                self.data[p] = min(s, v) if s is not None else v
        elif h is not None:
            for p, v in numpy.ndenumerate(self.data):
                self.data[p] = min(v, h) if v is not None else h

    def settomax(self, h):
        """Set all bins to the maximum of their current value and h.

        Args:
          h (Histogram2D or object): If `h` is a Histogram2D instance the
            function iterates over all bins of `h` and sets the corresponding
            bins of `self` to the maximum of the two values. Otherwise all bins
            of `self` are set to the maximum of their current value and `h`.

        """
        if isinstance(h, Histogram2D):
            for x, y, v in h.iteritems():
                p = self._getbin(x,y)
                s = self.data[p]
                self.data[p] = max(s, v) if s is not None else v
        elif h is not None:
            for p, v in numpy.ndenumerate(self.data):
                self.data[p] = max(v, h) if v is not None else h

    def contour(self, axes=None, convert=None, **kwargs):
        """Draw contour lines with matplotlib

        Args:
          axes (matplotlib.axes.Axes or None, optonal): Axes object to plot to.
            Uses ``matplotlib.pyplot.gca()`` if ``None``. Defaults to ``None``.
          convert (callable or None, optional): function that converts the
            elements of self.data to floats. If ``None``, the elements of
            ``self.data`` must be floats. Defaults to ``None``.
          **kwargs: all other keyword arguments are passed to
            ``matplotlib.axes.contour``.

        """
        if axes is None:
            axes = plt.gca()
        if convert is None:
            data = self.data
        else:
            data = numpy.vectorize(convert)(self.data)
        return axes.contour(self.xvals, self.yvals, data, **kwargs)

    def contourf(self, axes=None, convert=None, **kwargs):
        """Draw filled contours with matplotlib

        Args:
          axes (matplotlib.axes.Axes or None, optional): Axes object to plot to.
            Uses ``matplotlib.pyplot.gca()`` if None. Defaults to None.
          convert (callable or None, optional): function that converts the
            elements of ``self.data`` to floats. If ``None``, the elements of
            ``self.data`` must be floats. Defaults to ``None``.
          **kwargs: all other keyword arguments are passed to
            ``matplotlib.axes.contourf``.

        """
        if axes is None:
            axes = plt.gca()
        if convert is None:
            data = self.data
        else:
            data = numpy.vectorize(convert)(self.data)
        return axes.contourf(self.xvals, self.yvals, data, **kwargs)

    def pcolormesh(self, axes=None, convert=None, **kwargs):
        """Draw a colored mesh with matplotlib

        Args:
          axes (matplotlib.axes.Axes or None, optional): Axes object to plot to.
            Uses ``matplotlib.pyplot.gca()`` if ``None``. Defaults to ``None``.
          convert (callable or None, optional): function that converts the
            elements of ``self.data`` to floats. If ``None``, the elements of
            ``self.data`` must be floats. Defaults to ``None``.
          **kwargs: all other keyword arguments are passed to
            ``matplotlib.axes.pcolormesh``.

        """
        if axes is None:
            axes = plt.gca()
        if convert is None:
            data = self.data
        else:
            data = numpy.vectorize(convert)(self.data)
        return axes.pcolormesh(self.xbins, self.ybins, data, **kwargs)

    def __iadd__(self, h):
        if isinstance(h, Histogram2D):
            if not (numpy.array_equal(h.xbins, self.xbins) and \
                    numpy.array_equal(h.ybins, self.ybins)):
                raise ValueError('cannot add histograms with different bins')
            for p, v in numpy.ndenumerate(h.data):
                if self.data[p] is not None:
                    self.data[p] += v
        elif h is not None:
            for p, v in numpy.ndenumerate(self.data):
                if v is not None:
                    self.data[p] += h
        return self

    def __isub__(self, h):
        if isinstance(h, Histogram2D):
            if not (numpy.array_equal(h.xbins, self.xbins) and \
                    numpy.array_equal(h.ybins, self.ybins)):
                raise ValueError('cannot add histograms with different bins')
            for p, v in numpy.ndenumerate(h.data):
                if self.data[p] is not None:
                    self.data[p] -= v
        elif h is not None:
            for p, v in numpy.ndenumerate(self.data):
                if v is not None:
                    self.data[p] -= h
        return self

    def __imul__(self, h):
        if isinstance(h, Histogram2D):
            if not (numpy.array_equal(h.xbins, self.xbins) and \
                    numpy.array_equal(h.ybins, self.ybins)):
                raise ValueError('cannot add histograms with different bins')
            for p, v in numpy.ndenumerate(h.data):
                if self.data[p] is not None:
                    self.data[p] *= v
        elif h is not None:
            for p, v in numpy.ndenumerate(self.data):
                if v is not None:
                    self.data[p] *= h
        return self

    def __idiv__(self, h):
        if isinstance(h, Histogram2D):
            if not (numpy.array_equal(h.xbins, self.xbins) and \
                    numpy.array_equal(h.ybins, self.ybins)):
                raise ValueError('cannot add histograms with different bins')
            for p, v in numpy.ndenumerate(h.data):
                if self.data[p] is not None:
                    self.data[p] = _hdiv(self.data[p], v)
        elif h is not None:
            for p, v in numpy.ndenumerate(self.data):
                if v is not None:
                    self.data[p] = _hdiv(self.data[p], h)
        return self

    def __add__(self, h):
        result = self.copy()
        result += h
        return result

    def __sub__(self, h):
        result = self.copy()
        result -= h
        return result

    def __mul__(self, h):
        result = self.copy()
        result *= h
        return result

    def __div__(self, h):
        result = self.copy()
        result /= h
        return result

    def __radd__(self, h):
        result = self.copy()
        result += h
        return result

    def __rsub__(self, h):
        result = self.copy()
        result *= -1
        result += h
        return result

    def __rmul__(self, h):
        result = self.copy()
        result *= h
        return result

    def __rdiv__(self, h):
        result = Histogram2D(like=self)
        for p, v in numpy.ndenumerate(self.data):
            result.data[p] = _hdiv(h/v)
        return result


class Histogram1D:
    """Represents one-dimensional histograms.

    Args:
      xvals (list or None, optional): list of central x-values of the histogram
        bins. Derived from `xbins` if ``None``. Defaults to ``None``.
      xbins (list or None, optional): list of bin boundaries in x direction.
        Derived from `xvals` if ``None``. Defaults to ``None``.
      data (array of objects or None, optional): data members (z-values) of the
        histogram. If set to ``None`` a suitably dimensioned array of
        ``None`` values is created. Defaults to ``None``.
      like (Histogram1D or None, optional): If not ``None`` the central values
        and bin boundaries of another Histogram1D instance are used. This
        supercedes any values specified for `xvals` or `xbins`. The
        ``data`` member is not copied.

    Attributes:
      xvals (numpy array): list of central x-values of the histogram bins.
      xbins (numpy array): list of bin boundaries in x direction.
      data (numpy array): data members (z-values) of the histogram.

    Note:
      Histogram1D supports element-wise addition, subtraction, multiplication,
      and division with other Histogram1D instances as long as their
      `xvals`, `yvals`, `xbins` and `ybins` members are exactly equal. Addition,
      subtraction, multiplication or division by a constant value for bins
      is also supported.

    """
    def __init__(self, xvals=None, xbins=None, data=None, like=None):
        if like is not None:
            xvals = numpy.array(like.xvals)
            xbins = numpy.array(like.xbins)
        else:
            if xvals is None and xbins is None:
                raise TypeError('Either `like`, `xvals` or `xbins` must be set.')
            elif xvals is None:
                if not hasattr(xbins, '__len__'):
                    raise ValueError('Argument `xbins` must be iterable.')
                elif len(xbins) < 2:
                    raise ValueError('Argument `xbins` must have at least 2 entries.')
                xbins = numpy.array(xbins)
                xbins = numpy.array(xbins)
                xvals = 0.5*(xbins[1:] + xbins[:-1])
            else:
                if not hasattr(xvals, '__len__'):
                    raise ValueError('Argument `xvals` must be iterable.')
                elif len(xvals) < 2 and xbins is None:
                    raise ValueError('Argument `xvals` must have at least 2 entries or `xbins` must be given.')
                elif len(xvals) < 1:
                    raise ValueError('Argument `xvals` must have at least 1 entry.')
                xvals = numpy.array(xvals)
                if xbins is None:
                    xbins = 0.5*(xvals[1:] + xvals[:-1])
                    xbins = numpy.array([xvals[0]-(xbins[0]-xvals[0])] + \
                                            list(xbins) + \
                                            [xvals[-1]+(xvals[-1]-xbins[-1])])
                else:
                    xbins = numpy.array(xbins)
                    if len(xbins) != len(xvals)+1:
                        raise ValueError('Argument `xbins` has wrong length.')
                    for i, x in enumerate(xvals):
                        if x < xbins[i]:
                            raise ValueError('Bin centre outside of bin.')
                    if xvals[-1] > xbins[-1]:
                            raise ValueError('Bin centre outside of bin.')

        self.xvals = xvals
        self.xbins = xbins

        if data is None:
            self.data = numpy.empty((len(xvals),), dtype=object)
        else:
            if not hasattr(data, '__len__'):
                raise TypeError('`data` argument must be 1D-array.')
            if len(xvals) != len(data):
                raise ValueError('`data` argument has wrong dimension.')
            self.data = numpy.array(data)

    def copy(self):
        """Return an independent copy of self.

        """
        return Histogram1D(like=self, data=self.data)

    def _getbin(self, x):
        if x < self.xbins[0] or x > self.xbins[-1]:
            raise KeyError('Invalid key `'+repr(x)+'`.')
        return max(bisect.bisect_left(self.xbins, x) - 1, 0)

    def __getitem__(self, x):
        """Return the data value of a specific bin

        Args:
          x (float): An x coordinate.

        Returns:
          The data value of the bin which contains the coordinate `x`.

        Raises:
          KeyError: The coordinate `x` is outside the range of the histogram.

        """
        return self.data[self._getbin(x)]
        
    def __setitem__(self, x, val):
        """Set the data value of a specific bin

        Args:
          x (tuple): An x coordinate.
          val: A data value. The value of the bin containing the coordinate
            `x` will be set to `val`.

        Raises:
          KeyError: The coordinate `x` is outside the range of the histogram.

        """
        self.data[self._getbin(x)] = val

    def center(self, x):
        """Return the center of the bin to which x belongs.

        """
        return self.xvals[self._getbin(x)]

    def lb(self, x):
        """Return the lower bound of the bin to which x belongs.

        """
        return self.xbins[self._getbin(x)]

    def ub(self, x):
        """Return the upper bound of the bin to which x belongs.

        """
        return self.xbins[self._getbin(x)+1]
        
    def __iter__(self):
        """Iterate over the bin centres.

        Yields:
          x (float): The x-coordinate of the bin centre.

        """
        for x in self.xvals:
            yield x
    
    def iteritems(self):
        """Iterate over all elements of the histogram.

        Yields:
          tuple: ``(x, v)`` where `x` is the bin center and `v` the value
            associated with that bin.
        
        """
        for ix, x in enumerate(self.xvals):
            yield (x, self.data[ix])

    def max(self):
        """Return the largest data member in the histogram.

        Note:
          `None` values are ignored.

        """
        result = None
        for x, v in self.iteritems():
            if v is not None:
                result = v if result is None else max(v, result)
        return result

    def min(self):
        """Return the smalles data member in the histogram.

        Note:
          `None` values are ignored.

        """
        result = None
        for x, v in self.iteritems():
            if v is not None:
                result = v if result is None else min(v, result)
        return result
        
    def set(self, h):
        """Set all data values of the histogram.

        Args:
          h (Histogram1D or object): If `h` is a Histogram1D instance the
            function iterates over all bins of `h` and sets the corresponding
            bins of `self` to their values. Otherwise all bin values of
            `self` are set to `h`.

        """
        if isinstance(h, Histogram1D):
            for x, v in h.iteritems():
                self[x] = v
        else:
            self.data.fill(h)

    def settomin(self, h):
        """Set all bins to the minimum of their current value and h.

        Args:
          h (Histogram1D or object): If `h` is a Histogram1D instance the
            function iterates over all bins of `h` and sets the corresponding
            bins of `self` to the minimum of the two values. Otherwise all bins
            of `self` are set to the minimum of their current value and `h`.

        """
        if isinstance(h, Histogram1D):
            for x, v in h.iteritems():
                p = self._getbin(x)
                s = self.data[p]
                self.data[p] = min(s, v) if s is not None else v
        elif h is not None:
            for p, v in numpy.ndenumerate(self.data):
                self.data[p] = min(v, h) if v is not None else h

    def settomax(self, h):
        """Set all bins to the maximum of their current value and h.

        Args:
          h (Histogram1D or object): If `h` is a Histogram1D instance the
            function iterates over all bins of `h` and sets the corresponding
            bins of `self` to the maximum of the two values. Otherwise all bins
            of `self` are set to the maximum of their current value and `h`.

        """
        if isinstance(h, Histogram1D):
            for x, v in h.iteritems():
                p = self._getbin(x)
                s = self.data[p]
                self.data[p] = max(s, v) if s is not None else v
        elif h is not None:
            for p, v in numpy.ndenumerate(self.data):
                self.data[p] = max(v, h) if v is not None else h

    def plot(self, *args, **kwargs):
        """Plot the histogram with matplotlib.

        Args:
          *args: all positional arguments are passed to matplotlib.axes.Axes.plot
          axes (matplotlib.axes.Axes or None, optional): Axes object to plot to.
            Uses ``matplotlib.pyplot.gca()`` if ``None``. Defaults to ``None``.
          convert (callable or None, optional): function that converts the
            elements of ``self.data`` to floats. If ``None``, the elements of
            ``self.data`` must be floats or have the attribute ``mean``.
          **kwargs: all other keyword arguments are passed to
            ``matplotlib.axes.Axes.plot``.

        """
        axes = kwargs.get('axes', None)
        convert = kwargs.get('convert', None)
        if axes is None:
            axes = plt.gca()
        if convert is None:
            convert = lambda x: x.mean if \
                      hasattr(x, 'mean') and hasattr(x, 'sdev') else x
        if len(self.data) == 0:
            return
        dat = [(convert(d) if d is not None else None) for d in self.data]
        if kwargs.get('drawstyle', None) == 'steps':
            dat = [dat[0]]+dat
            xvals = self.xbins
        else:
            xvals = self.xvals
        kwargs.pop('axes', None)
        kwargs.pop('convert', None)
        return axes.plot(xvals, dat, *args, **kwargs)

    def errorbars(self, *args, **kwargs):
        """Plot errorbars with matplotlib.

        Args:
          *args: all positional arguments are passed to
            ``matplotlib.axes.Axes.errorbars``
          axes (matplotlib.axes.Axes or None, optional): Axes object to plot to.
            Uses ``matplotlib.pyplot.gca()`` if ``None``. Defaults to ``None``.
          convert (callable or None, optional): function that converts the
            elements of ``self.data`` to ``GVar`` objects. If ``None``, the
            elements of ``self.data`` must be be convertible by the ``GVar``
            constructor.
          **kwargs: all other keyword arguments are passed to
            ``matplotlib.axes.Axes.errorbars``.

        """
        axes = kwargs.get('axes', None)
        convert = kwargs.get('convert', None)
        if axes is None:
            axes = plt.gca()
        if convert is None:
            convert = lambda x: GVar(x)
        if len(self.data) == 0:
            return
        means = []
        sdevs = []
        xvals = []
        for x, d in izip(self.xvals, self.data):
            if d is not None:
                gd = convert(d)
                xvals.append(x)
                means.append(gd.mean)
                sdevs.append(gd.sdev)
        kwargs['yerr'] = sdevs
        kwargs['fmt'] = None
        kwargs.pop('axes', None)
        kwargs.pop('convert', None)
        return axes.errorbar(xvals, means, *args, **kwargs)

    def setxlim(self, axes=None):
        """Set x limits of the current plot.

        Args:
          axes (matplotlib.axes.Axes or None): Axes object whose limits are
            to be set. Uses ``matplotlib.pyplot.gca()`` if ``None``. Defaults to
            ``None``.

        """
        if axes is None:
            axes = plt.gca()
        return axes.set_xlim((self.xbins[0], self.xbins[-1]))

    def __iadd__(self, h):
        if isinstance(h, Histogram1D):
            if not numpy.array_equal(h.xbins, self.xbins):
                raise ValueError('cannot add histograms with different bins')
            for p, v in numpy.ndenumerate(h.data):
                if self.data[p] is not None:
                    self.data[p] += v
        elif h is not None:
            for p, v in numpy.ndenumerate(self.data):
                if v is not None:
                    self.data[p] += h
        return self

    def __isub__(self, h):
        if isinstance(h, Histogram1D):
            if not numpy.array_equal(h.xbins, self.xbins):
                raise ValueError('cannot add histograms with different bins')
            for p, v in numpy.ndenumerate(h.data):
                if self.data[p] is not None:
                    self.data[p] -= v
        elif h is not None:
            for p, v in numpy.ndenumerate(self.data):
                if v is not None:
                    self.data[p] -= h
        return self

    def __imul__(self, h):
        if isinstance(h, Histogram1D):
            if not numpy.array_equal(h.xbins, self.xbins):
                raise ValueError('cannot add histograms with different bins')
            for p, v in numpy.ndenumerate(h.data):
                if self.data[p] is not None:
                    self.data[p] *= v
        elif h is not None:
            for p, v in numpy.ndenumerate(self.data):
                if v is not None:
                    self.data[p] *= h
        return self

    def __idiv__(self, h):
        if isinstance(h, Histogram1D):
            if not numpy.array_equal(h.xbins, self.xbins):
                raise ValueError('cannot add histograms with different bins')
            for p, v in numpy.ndenumerate(h.data):
                if self.data[p] is not None:
                    self.data[p] = _hdiv(self.data[p], v)
        elif h is not None:
            for p, v in numpy.ndenumerate(self.data):
                if v is not None:
                    self.data[p] = _hdiv(self.data[p], h)
        return self

    def __add__(self, h):
        result = self.copy()
        result += h
        return result

    def __sub__(self, h):
        result = self.copy()
        result -= h
        return result

    def __mul__(self, h):
        result = self.copy()
        result *= h
        return result

    def __div__(self, h):
        result = self.copy()
        result /= h
        return result

    def __radd__(self, h):
        result = self.copy()
        result += h
        return result

    def __rsub__(self, h):
        result = self.copy()
        result *= -1
        result += h
        return result

    def __rmul__(self, h):
        result = self.copy()
        result *= h
        return result

    def __rdiv__(self, h):
        result = Histogram2D(like=self)
        for p, v in numpy.ndenumerate(self.data):
            result.data[p] = _hdiv(h, v)
        return result


def cumulatedHistogram(h, upper=False, const=None):
    """Compute the cumulated distribution.

    Args:
      h (Histogram1D): the histogram to cumulate
      upper (bool, optional): If ``True``, the histogram is cumulated from the
        upper end. Defaults to ``False``.
      const (optional): Constant to be added to all bins of the resulting
        histogram. Defaults to ``None``, in which case nothing is added.

    Returns:
      Histogram1D: The cumulated histogram.

    Note:
      If ``hc = cumulated(h, upper=False, const=c)`` then 
      ``hc.data[i] = sum(h.data[:i+1]) + c`` for all `i` in
      ``range(len(h.data))``.
      If ``hc = cumulated(h, upper=True, const=c)`` then 
      ``hc.data[i] = sum(h.data[i:]) + c`` for all `i` in ``range(len(h.data))``.

    """
    if not isinstance(h, Histogram1D):
        raise ValueError('Argument must be Histogram1D instance.')
    result = Histogram1D(like=h)
    if len(result.data) <= 0:
        return result
    if upper:
        if const is not None:
            result.data[-1] = const + h.data[-1]
        else:
            result.data[-1] = h.data[-1]
        for i in reversed(range(len(result.data)-1)):
            result.data[i] = h.data[i] + result.data[i+1]
    else:
        if const is not None:
            result.data[0] = const + h.data[0]
        else:
            result.data[0] = h.data[0]
        for i in range(1, len(result.data)):
            result.data[i] = h.data[i] + result.data[i-1]
    return result

def densityHistogram(h):
    """Divide all bin values by the bin width (Histogram1D) or bin area
    (Histogram2D).

    Args:
      h (Histogram1D or Histogram2D): the histogram to convert. `h` is not
        modified.

    Returns:
      Histogram1D or Histogram2D: the converted histogram.

    """
    if isinstance(h, Histogram1D):
        result = Histogram1D(like=h)
        if len(result.data) <= 0:
            return result
        for i in range(len(result.xbins)-1):
            result.data[i] = h.data[i]/(h.xbins[i+1]-h.xbins[i])
        return result
    elif isinstance(h, Histogram2D):
        result = Histogram1D(like=h)
        if len(result.data) <= 0:
            return result
        for i in range(len(result.xbins)-1):
            for j in range(len(result.ybins)-1):
                result.data[j,i] = h.data[j,i]/ \
                    ((h.xbins[i+1]-h.xbins[i])*(h.ybins[j+1]-h.ybins[j]))
    else:
        raise ValueError('Argument must be Histogram1D or Histogram2D instance.')


def _split(nx, kx):
    dx = nx // kx
    if dx <= 0:
        raise ValueError(
            'Cannot split {0:i} elements into {1:i} blocks.'.format(nx, kx))
    rx = nx % kx
    ranges = []
    for sid in range(kx):
        ix = sid*(dx+1) if sid < rx else rx*(dx+1)+(sid-rx)*dx
        jx = ix + dx + 1 if sid < rx else ix + dx
        ranges.append((ix, jx))
    return ranges


class SplitHistogram1D:
    """Split 1D histogram into (roughly) equal-sized sub-histograms.

    Note:
      This class is typically used for parallelisation together with
      :py:class:`~myFitter.ParallelFunction` and
      :py:meth:`Profiler1D.scan() <myFitter.Profiler1D.scan>`::

        profiler = Profiler1D(...)
        hist = Histogram1D(...)
        splithist = SplitHistogram1D(hist, 5)
        splithist.subhists = \\
            ParallelFunction(profiler.scan, batchsize=1)(splithist.subhists)
        hist = splithist.join()

    Args:
      h (Histogram1D): The histogram to split.
      xblocks (int): The number of blocks in the x-direction.

    Attributes:
      subhists (list of Histogram1D): The list of sub-histograms.

    """
    def __init__(self, h, xblocks):
        self._xblocks = xblocks
        self._nx = len(h.xvals)
        xranges = _split(self._nx, self._xblocks)
        self.subhists = []
        for i, j in xranges:
            self.subhists.append(Histogram1D(
                xvals=h.xvals[i:j],
                xbins=h.xbins[i:j+1],
                data=h.data[i:j]))
        self._xbounds = [b.xbins[0] for b in self.subhists]
        self._xbounds.append(self.subhists[-1].xbins[-1])

    def find(self, x):
        """Find the index of the sub-histogram containing `x`.

        Args:
          x (float): x-value to search for.

        Raises:
          KeyError: None of the sub-histograms contains the x-value `x`.

        Returns:
          int: The index of the sub-histogram (in
            :py:attr:`~myFitter.SplitHistogram1D.subhists`) containing the
            x-value `x`.
            
        """
        if x < self._xbounds[0] or x > self._xbounds[-1]:
            raise KeyError('Invalid key `'+repr(x)+'`.')
        return max(bisect.bisect_left(self._xbounds, x) - 1, 0)

    def subhist(self, x):
        """Find the sub-histogram containing `x`.

        Args:
          x (float): x-value to search for.

        Raises:
          KeyError: None of the sub-histograms contains the x-value `x`.

        Returns:
          Histogram1D: The sub-histogram containing the x-value `x`.
            
        """
        return self.subhists[self.find(x)]

    def __getitem__(self, x):
        """Return the data value of a specific bin

        Args:
          x (float): An x coordinate.

        Returns:
          The data value of the bin which contains the coordinate `x`.

        Raises:
          KeyError: The coordinate `x` is outside the range of all
            sub-histograms.

        """
        return self.subhist(x)[x]

    def join(self):
        """Join all sub-histograms into a single histogram.

        Returns:
          Histogram1D: The joined histogram.

        """
        nx = self._nx
        xvals = numpy.empty(nx)
        xbins = numpy.empty(nx+1)
        data = numpy.empty(nx, dtype=object)
        i = 0
        xbins[0] = self.subhists[0].xbins[0]
        for block in self.subhists:
            for xval, xbin, v in zip(block.xvals, block.xbins[1:], block.data):
                xvals[i] = xval
                xbins[i+1] = xbin
                data[i] = v
                i = i + 1                
        return Histogram1D(xvals=xvals, xbins=xbins, data=data)


class SplitHistogram2D:
    """Split 2D histogram into (roughly) equal-sized sub-histograms.

    Note:
      This class is typically used for parallelisation together with
      :py:class:`~myFitter.ParallelFunction` and
      :py:meth:`Profiler2D.scan() <myFitter.Profiler2D.scan>`::

        profiler = Profiler2D(...)
        hist = Histogram2D(...)
        splithist = SplitHistogram2D(hist, 2, 3)
        splithist.subhists = \\
            ParallelFunction(profiler.scan, batchsize=1)(splithist.subhists)
        hist = splithist.join()

    Args:
      h (Histogram2D): The histogram to split.
      xblocks (int): The number of blocks in the x-direction.
      yblocks (int): The number of blocks in the y-direction.

    Attributes:
      subhists (list of Histogram2D): The list of sub-histograms.


    """
    def __init__(self, h, xblocks, yblocks):
        self._xblocks = xblocks
        self._yblocks = yblocks
        self._nx = len(h.xvals)
        self._ny = len(h.yvals)
        xranges = _split(self._nx, self._xblocks)
        yranges = _split(self._ny, self._yblocks)
        self.subhists = []
        for iy, jy in yranges:
            for ix, jx in xranges:
                self.subhists.append(Histogram2D(
                    xvals=h.xvals[ix:jx],
                    xbins=h.xbins[ix:jx+1],
                    yvals=h.yvals[iy:jy],
                    ybins=h.ybins[iy:jy+1],
                    data=h.data[iy:jy, ix:jx]))
        self._xbounds = [self.subhists[ib].xbins[0] \
                         for ib in xrange(self._xblocks)]
        self._xbounds.append(self.subhists[-1].xbins[-1])
        self._ybounds = [self.subhists[ib*self._xblocks].ybins[0] \
                         for ib in xrange(self._yblocks)]
        self._ybounds.append(self.subhists[-1].ybins[-1])


    def find(self, x, y):
        """Find the index of the sub-histogram containing ``(x, y)``.

        Args:
          x (float): x-coordiante of the point to find.
          y (float): y-coordiante of the point to find.

        Raises:
          KeyError: None of the sub-histograms contains the point ``(x, y)``.

        Returns:
          int: The index of the sub-histogram (in
            :py:attr:`~myFitter.SplitHistogram1D.subhists`) containing the
            point ``(x, y)``.
            
        """
        if x < self._xbounds[0] or x > self._xbounds[-1]:
            raise KeyError('Invalid key `'+repr(x)+'`.')
        ix = max(bisect.bisect_left(self._xbounds, x) - 1, 0)
        if y < self._ybounds[0] or y > self._ybounds[-1]:
            raise KeyError('Invalid key `'+repr(y)+'`.')
        iy = max(bisect.bisect_left(self._ybounds, y) - 1, 0)
        return iy*self._xblocks + ix

    def subhist(self, x, y):
        """Find the sub-histogram containing ``(x, y)``.

        Args:
          x (float): x-coordinate of the point to search for.
          y (float): y-coordinate of the point to search for.

        Raises:
          KeyError: None of the sub-histograms contains the point ``(x, y)``.

        Returns:
          Histogram2D: The sub-histogram containing the point ``(x, y)``.
            
        """
        return self.subhists[self.find(x, y)]

    def __getitem__(self, p):
        """Return the data value of a specific bin

        Args:
          p (tuple of float): A pair of x and y coordinates.

        Returns:
          The data value of the bin which contains the point `p`.

        Raises:
          KeyError: The coordinates `p` are outside the range of all
            sub-histograms.

        """
        x, y = p
        return self.subhist(x, y)[p]

    def join(self):
        """Join all sub-histograms into a single histogram.

        Returns:
          Histogram2D: The joined histogram.

        """
        nx = self._nx
        ny = self._ny
        nbx = self._xblocks
        nby = self._yblocks
        xvals = numpy.empty(nx)
        xbins = numpy.empty(nx+1)
        yvals = numpy.empty(ny)
        ybins = numpy.empty(ny+1)
        data = numpy.empty((ny, nx), dtype=object)

        xbins[0] = self.subhists[0].xbins[0]
        ybins[0] = self.subhists[0].ybins[0]
        iy = 0
        for iby in xrange(nby):
            block = self.subhists[iby*nbx]
            dy = len(block.yvals)
            yvals[iy:iy+dy] = block.yvals
            ybins[iy+1:iy+1+dy] = block.ybins[1:]
            iy = iy + dy
        ix = 0
        for ibx in xrange(nbx):
            block = self.subhists[ibx]
            dx = len(block.xvals)
            xvals[ix:ix+dx] = block.xvals
            xbins[ix+1:ix+1+dx] = block.xbins[1:]
            ix = ix + dx
        iy = 0
        for iby in xrange(nby):
            block = self.subhists[iby*nbx]
            dy = len(block.yvals)
            ix = 0
            for ibx in xrange(nbx):
                block = self.subhists[iby*nbx+ibx]
                dx = len(block.xvals)
                data[iy:iy+dy, ix:ix+dx] = block.data
                ix = ix + dx
            iy = iy + dy

        return Histogram2D(xvals=xvals, xbins=xbins,
                           yvals=yvals, ybins=ybins, data=data)
