import unittest
import numpy as np
from myFitter import *


def eq(a, b, rprec=1e-10, aprec=1e-10):
    if hasattr(a, 'shape'):
        if hasattr(b, 'shape') and a.shape == b.shape:
            for ai, bi in zip(a.flat, b.flat):
                diff = abs(ai-bi)
                avg = 0.5*(abs(ai)+abs(bi))
                if avg >= aprec and diff >= rprec*avg:
                    return False
                return True
        else:
            raise ValueError('Arrays a and b must have the same shape.')
    else:
        diff = abs(a-b)
        avg = 0.5*(abs(a)+abs(b))
        return avg < aprec or diff < rprec*avg
            

class HistogramsTest(unittest.TestCase):
    def setUp(self):
        pass

    def test_histogram1d(self):
        xbins = np.linspace(0.0, 10.0, 11)
        xvals = np.linspace(0.5, 9.5, 10)
        data  = np.linspace(0.0, 9.0, 10)

        hist = Histogram1D(xbins=xbins, data=data)
        self.assertTrue(eq(hist.xvals, xvals))
        
        hist = Histogram1D(xvals=xvals, data=data)
        self.assertTrue(eq(hist.xbins, xbins))

        hist = Histogram1D(xvals=xvals, xbins=xbins, data=data)
        splithist = SplitHistogram1D(hist, 3)
        self.assertEqual(len(splithist.subhists), 3)
        self.assertTrue(eq(splithist.subhists[0].xvals, xvals[0:4]))
        self.assertTrue(eq(splithist.subhists[0].data, data[0:4]))
        self.assertTrue(eq(splithist.subhists[0].xbins, xbins[0:5]))
        self.assertTrue(eq(splithist.subhists[-1].xvals, xvals[7:10]))
        self.assertTrue(eq(splithist.subhists[-1].data, data[7:10]))
        self.assertTrue(eq(splithist.subhists[-1].xbins, xbins[7:11]))

        self.assertEqual(splithist.find(3.3), 0)
        self.assertEqual(splithist.find(4.1), 1)
        self.assertEqual(splithist.find(9.6), 2)

    def test_histogram2d(self):
        xbins = np.linspace(0.0, 10.0, 11)
        xvals = np.linspace(0.5, 9.5, 10)
        ybins = np.linspace(0.0,  5.0, 6)
        yvals = np.linspace(0.5,  4.5, 5)
        data = np.empty((5, 10))
        for iy in xrange(5):
            for ix in xrange(10):
                data[iy, ix] = 10.0*iy + ix

        hist = Histogram2D(xbins=xbins, ybins=ybins, data=data)
        self.assertTrue(eq(hist.xvals, xvals))
        self.assertTrue(eq(hist.yvals, yvals))
        
        hist = Histogram2D(xvals=xvals, yvals=yvals, data=data)
        self.assertTrue(eq(hist.xbins, xbins))
        self.assertTrue(eq(hist.ybins, ybins))

        hist = Histogram2D(xvals=xvals, xbins=xbins,
                           yvals=yvals, ybins=ybins, data=data)
        splithist = SplitHistogram2D(hist, 3, 2)
        self.assertEqual(len(splithist.subhists), 3*2)

        self.assertTrue(eq(splithist.subhists[0].xvals, xvals[0:4]))
        self.assertTrue(eq(splithist.subhists[0].xbins, xbins[0:5]))
        self.assertTrue(eq(splithist.subhists[0].yvals, yvals[0:3]))
        self.assertTrue(eq(splithist.subhists[0].ybins, ybins[0:4]))
        self.assertTrue(eq(splithist.subhists[0].data, data[0:3,0:4]))

        self.assertTrue(eq(splithist.subhists[3].xvals, xvals[0:4]))
        self.assertTrue(eq(splithist.subhists[3].xbins, xbins[0:5]))
        self.assertTrue(eq(splithist.subhists[3].yvals, yvals[3:5]))
        self.assertTrue(eq(splithist.subhists[3].ybins, ybins[3:6]))
        self.assertTrue(eq(splithist.subhists[3].data, data[3:5,0:4]))

        self.assertTrue(eq(splithist.subhists[-1].xvals, xvals[7:10]))
        self.assertTrue(eq(splithist.subhists[-1].xbins, xbins[7:11]))
        self.assertTrue(eq(splithist.subhists[-1].yvals, yvals[3:5]))
        self.assertTrue(eq(splithist.subhists[-1].ybins, ybins[3:6]))
        self.assertTrue(eq(splithist.subhists[-1].data, data[3:5,7:10]))

        self.assertEqual(splithist.find(3.3, 2.6), 0*3+0)
        self.assertEqual(splithist.find(4.1, 3.6), 1*3+1)
        self.assertEqual(splithist.find(9.6, 3.6), 1*3+2)
        
